// Copyright 2018 The Chromium Authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#include "third_party/blink/renderer/core/trustedtypes/trusted_type_policy_factory.h"

#include "third_party/blink/public/mojom/web_feature/web_feature.mojom-blink.h"
#include "third_party/blink/renderer/bindings/core/v8/script_value.h"
#include "third_party/blink/renderer/bindings/core/v8/v8_trusted_html.h"
#include "third_party/blink/renderer/bindings/core/v8/v8_trusted_script.h"
#include "third_party/blink/renderer/bindings/core/v8/v8_trusted_script_url.h"
#include "third_party/blink/renderer/bindings/core/v8/v8_trusted_url.h"
#include "third_party/blink/renderer/core/dom/document.h"
#include "third_party/blink/renderer/core/frame/csp/content_security_policy.h"
#include "third_party/blink/renderer/core/frame/local_frame.h"
#include "third_party/blink/renderer/core/trustedtypes/trusted_type_policy.h"
#include "third_party/blink/renderer/core/trustedtypes/trusted_types_util.h"
#include "third_party/blink/renderer/platform/bindings/exception_state.h"
#include "third_party/blink/renderer/platform/instrumentation/use_counter.h"
#include "third_party/blink/renderer/platform/wtf/text/string_hash.h"

namespace blink {

TrustedTypePolicy* TrustedTypePolicyFactory::createPolicy(
    const String& policy_name,
    const TrustedTypePolicyOptions* policy_options,
    ExceptionState& exception_state) {
  if (!GetExecutionContext()) {
    exception_state.ThrowDOMException(DOMExceptionCode::kInvalidStateError,
                                      "The document is detached.");
    return nullptr;
  }
  UseCounter::Count(GetExecutionContext(),
                    WebFeature::kTrustedTypesCreatePolicy);
  if (RuntimeEnabledFeatures::TrustedDOMTypesEnabled(GetExecutionContext()) &&
      GetExecutionContext()->GetContentSecurityPolicy() &&
      !GetExecutionContext()
           ->GetContentSecurityPolicy()
           ->AllowTrustedTypePolicy(policy_name)) {
    exception_state.ThrowTypeError("Policy " + policy_name + " disallowed.");
    return nullptr;
  }
  // TODO(orsibatiz): After policy naming rules are estabilished, check for the
  // policy_name to be according to them.
  if (policy_map_.Contains(policy_name)) {
    exception_state.ThrowTypeError("Policy with name " + policy_name +
                                   " already exists.");
    return nullptr;
  }
  if (policy_name == "default") {
    UseCounter::Count(GetExecutionContext(),
                      WebFeature::kTrustedTypesDefaultPolicyUsed);
  }
  auto* policy = MakeGarbageCollected<TrustedTypePolicy>(
      policy_name, const_cast<TrustedTypePolicyOptions*>(policy_options));
  policy_map_.insert(policy_name, policy);
  return policy;
}

TrustedTypePolicy* TrustedTypePolicyFactory::defaultPolicy() const {
  return policy_map_.at("default");
}

TrustedTypePolicyFactory::TrustedTypePolicyFactory(ExecutionContext* context)
    : ContextClient(context),
      empty_html_(MakeGarbageCollected<TrustedHTML>("")) {
  UseCounter::Count(context, WebFeature::kTrustedTypesEnabled);
}

Vector<String> TrustedTypePolicyFactory::getPolicyNames() const {
  Vector<String> policyNames;
  for (const String name : policy_map_.Keys()) {
    policyNames.push_back(name);
  }
  return policyNames;
}

const WrapperTypeInfo*
TrustedTypePolicyFactory::GetWrapperTypeInfoFromScriptValue(
    ScriptState* script_state,
    const ScriptValue& script_value) {
  v8::Local<v8::Value> value = script_value.V8Value();
  if (value.IsEmpty() || !value->IsObject() ||
      !V8DOMWrapper::IsWrapper(script_state->GetIsolate(), value))
    return nullptr;
  v8::Local<v8::Object> object = value.As<v8::Object>();
  return ToWrapperTypeInfo(object);
}

bool TrustedTypePolicyFactory::isHTML(ScriptState* script_state,
                                      const ScriptValue& script_value) {
  const WrapperTypeInfo* wrapper_type_info =
      GetWrapperTypeInfoFromScriptValue(script_state, script_value);
  return wrapper_type_info &&
         wrapper_type_info->Equals(V8TrustedHTML::GetWrapperTypeInfo());
}

bool TrustedTypePolicyFactory::isScript(ScriptState* script_state,
                                        const ScriptValue& script_value) {
  const WrapperTypeInfo* wrapper_type_info =
      GetWrapperTypeInfoFromScriptValue(script_state, script_value);
  return wrapper_type_info &&
         wrapper_type_info->Equals(V8TrustedScript::GetWrapperTypeInfo());
}

bool TrustedTypePolicyFactory::isScriptURL(ScriptState* script_state,
                                           const ScriptValue& script_value) {
  const WrapperTypeInfo* wrapper_type_info =
      GetWrapperTypeInfoFromScriptValue(script_state, script_value);
  return wrapper_type_info &&
         wrapper_type_info->Equals(V8TrustedScriptURL::GetWrapperTypeInfo());
}

bool TrustedTypePolicyFactory::isURL(ScriptState* script_state,
                                     const ScriptValue& script_value) {
  const WrapperTypeInfo* wrapper_type_info =
      GetWrapperTypeInfoFromScriptValue(script_state, script_value);
  return wrapper_type_info &&
         wrapper_type_info->Equals(V8TrustedURL::GetWrapperTypeInfo());
}

TrustedHTML* TrustedTypePolicyFactory::emptyHTML() const {
  return empty_html_.Get();
}

const struct {
  const char* element;
  const char* property;
  const char* element_namespace;
  SpecificTrustedType type;
  // We use this table for both attributes and properties. Most are both,
  // because DOM "reflects" the attributes onto their specified poperties.
  // We use is_not_* so that the default value initialization (false) will
  // match the common case and we only need to explicitly provide the uncommon
  // values.
  bool is_not_property : 1;
  bool is_not_attribute : 1;
} kTypeTable[] = {
    {"a", "href", nullptr, SpecificTrustedType::kTrustedURL},
    {"area", "href", nullptr, SpecificTrustedType::kTrustedURL},
    {"audio", "src", nullptr, SpecificTrustedType::kTrustedURL},
    {"base", "href", nullptr, SpecificTrustedType::kTrustedURL},
    {"button", "formAction", nullptr, SpecificTrustedType::kTrustedURL},
    {"embed", "src", nullptr, SpecificTrustedType::kTrustedScriptURL},
    {"form", "action", nullptr, SpecificTrustedType::kTrustedURL},
    {"frame", "src", nullptr, SpecificTrustedType::kTrustedURL},
    {"iframe", "src", nullptr, SpecificTrustedType::kTrustedURL},
    {"iframe", "srcdoc", nullptr, SpecificTrustedType::kTrustedHTML},
    {"img", "src", nullptr, SpecificTrustedType::kTrustedURL},
    {"input", "formAction", nullptr, SpecificTrustedType::kTrustedURL},
    {"input", "src", nullptr, SpecificTrustedType::kTrustedURL},
    {"link", "href", nullptr, SpecificTrustedType::kTrustedURL},
    {"object", "codeBase", nullptr, SpecificTrustedType::kTrustedScriptURL},
    {"object", "data", nullptr, SpecificTrustedType::kTrustedScriptURL},
    {"script", "innerText", nullptr, SpecificTrustedType::kTrustedScript, false,
     true},
    {"script", "src", nullptr, SpecificTrustedType::kTrustedScriptURL},
    {"script", "text", nullptr, SpecificTrustedType::kTrustedScript, false,
     true},
    {"script", "textContent", nullptr, SpecificTrustedType::kTrustedScript,
     false, true},
    {"source", "src", nullptr, SpecificTrustedType::kTrustedURL},
    {"track", "src", nullptr, SpecificTrustedType::kTrustedURL},
    {"video", "src", nullptr, SpecificTrustedType::kTrustedURL},
    {"*", "innerHTML", nullptr, SpecificTrustedType::kTrustedHTML, false, true},
    {"*", "outerHTML", nullptr, SpecificTrustedType::kTrustedHTML, false, true},
    {"*", "on*", nullptr, SpecificTrustedType::kTrustedScript, true, false},
};

// Does a type table entry match a property?
// (Properties are evaluated by JavaScript and are case-sensitive.)
bool EqualsProperty(decltype(*kTypeTable)& left,
                    const String& tag,
                    const String& attr,
                    const String& ns) {
  DCHECK_EQ(tag.LowerASCII(), tag);
  return (left.element == tag || !strcmp(left.element, "*")) &&
         (left.property == attr ||
          (!strcmp(left.property, "on*") && attr.StartsWith("on"))) &&
         left.element_namespace == ns && !left.is_not_property;
}

// Does a type table entry match an attribute?
// (Attributes get queried by calling acecssor methods on the DOM. These are
//  case-insensitivem, because DOM.)
bool EqualsAttribute(decltype(*kTypeTable)& left,
                     const String& tag,
                     const String& attr,
                     const String& ns) {
  DCHECK_EQ(tag.LowerASCII(), tag);
  return (left.element == tag || !strcmp(left.element, "*")) &&
         (String(left.property).LowerASCII() == attr.LowerASCII() ||
          (!strcmp(left.property, "on*") && attr.StartsWith("on"))) &&
         left.element_namespace == ns && !left.is_not_attribute;
}

String getTrustedTypeName(SpecificTrustedType type) {
  switch (type) {
    case SpecificTrustedType::kTrustedHTML:
      return "TrustedHTML";
    case SpecificTrustedType::kTrustedScript:
      return "TrustedScript";
    case SpecificTrustedType::kTrustedScriptURL:
      return "TrustedScriptURL";
    case SpecificTrustedType::kTrustedURL:
      return "TrustedURL";
    case SpecificTrustedType::kNone:
      return String();
  }
}

typedef bool (*PropertyEqualsFn)(decltype(*kTypeTable)&,
                                 const String&,
                                 const String&,
                                 const String&);

String FindTypeInTypeTable(const String& tagName,
                           const String& propertyName,
                           const String& elementNS,
                           PropertyEqualsFn equals) {
  SpecificTrustedType type = SpecificTrustedType::kNone;
  for (auto* it = std::cbegin(kTypeTable); it != std::cend(kTypeTable); it++) {
    if ((*equals)(*it, tagName, propertyName, elementNS)) {
      type = it->type;
      break;
    }
  }
  return getTrustedTypeName(type);
}

String TrustedTypePolicyFactory::getPropertyType(
    const String& tagName,
    const String& propertyName,
    const String& elementNS) const {
  return FindTypeInTypeTable(tagName.LowerASCII(), propertyName, elementNS,
                             &EqualsProperty);
}

String TrustedTypePolicyFactory::getAttributeType(
    const String& tagName,
    const String& attributeName,
    const String& tagNS,
    const String& attributeNS) const {
  return FindTypeInTypeTable(tagName.LowerASCII(), attributeName, tagNS,
                             &EqualsAttribute);
}

String TrustedTypePolicyFactory::getPropertyType(
    const String& tagName,
    const String& propertyName) const {
  return getPropertyType(tagName, propertyName, String());
}

String TrustedTypePolicyFactory::getAttributeType(
    const String& tagName,
    const String& attributeName) const {
  return getAttributeType(tagName, attributeName, String(), String());
}

String TrustedTypePolicyFactory::getAttributeType(const String& tagName,
                                                  const String& attributeName,
                                                  const String& tagNS) const {
  return getAttributeType(tagName, attributeName, tagNS, String());
}

ScriptValue TrustedTypePolicyFactory::getTypeMapping(
    ScriptState* script_state) const {
  return getTypeMapping(script_state, String());
}

ScriptValue TrustedTypePolicyFactory::getTypeMapping(ScriptState* script_state,
                                                     const String& ns) const {
  // Create three-deep dictionary of properties, like so:
  // {tagname: { ["attributes"|"properties"]: { attribute: type }}}

  if (!ns.IsEmpty())
    return ScriptValue();

  v8::HandleScope handle_scope(script_state->GetIsolate());
  v8::Local<v8::Object> top = v8::Object::New(script_state->GetIsolate());
  v8::Local<v8::Object> properties;
  v8::Local<v8::Object> attributes;
  const char* element = nullptr;
  for (const auto& iter : kTypeTable) {
    if (properties.IsEmpty() || !element || strcmp(iter.element, element)) {
      element = iter.element;
      v8::Local<v8::Object> middle =
          v8::Object::New(script_state->GetIsolate());
      top->Set(script_state->GetContext(),
               V8String(script_state->GetIsolate(), iter.element), middle)
          .Check();
      properties = v8::Object::New(script_state->GetIsolate());
      middle
          ->Set(script_state->GetContext(),
                V8String(script_state->GetIsolate(), "properties"), properties)
          .Check();
      attributes = v8::Object::New(script_state->GetIsolate());
      middle
          ->Set(script_state->GetContext(),
                V8String(script_state->GetIsolate(), "attributes"), attributes)
          .Check();
    }
    if (!iter.is_not_property) {
      properties
          ->Set(script_state->GetContext(),
                V8String(script_state->GetIsolate(), iter.property),
                V8String(script_state->GetIsolate(),
                         getTrustedTypeName(iter.type)))
          .Check();
    }
    if (!iter.is_not_attribute) {
      attributes
          ->Set(script_state->GetContext(),
                V8String(script_state->GetIsolate(), iter.property),
                V8String(script_state->GetIsolate(),
                         getTrustedTypeName(iter.type)))
          .Check();
    }
  }

  return ScriptValue(script_state, top);
}

void TrustedTypePolicyFactory::CountTrustedTypeAssignmentError() {
  if (!hadAssignmentError) {
    UseCounter::Count(GetExecutionContext(),
                      WebFeature::kTrustedTypesAssignmentError);
    hadAssignmentError = true;
  }
}

void TrustedTypePolicyFactory::Trace(blink::Visitor* visitor) {
  ScriptWrappable::Trace(visitor);
  ContextClient::Trace(visitor);
  visitor->Trace(empty_html_);
  visitor->Trace(policy_map_);
}

}  // namespace blink
